/*
** Copyright (C) Red Hat, Inc. 2005-2011
**
** This program is free software; you can redistribute it and/or modify it
** under the terms of the GNU General Public License version 2 as
** published by the Free Software Foundation.
**
** This program is distributed in the hope that it will be useful, but
** WITHOUT ANY WARRANTY; without even the implied warranty of
** MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
** General Public License for more details.
**
** You should have received a copy of the GNU General Public License
** along with this program; see the file COPYING. If not, write to the
** Free Software Foundation, Inc., 675 Mass Ave, Cambridge,
** MA 02139, USA.
*/

/*
 * Author: Stanko Kupcevic <kupcevic@redhat.com>
 */

#ifndef __CONGA_SOCKET_H
#define __CONGA_SOCKET_H

#include "counting_auto_ptr.h"

#include "String.h"
#include <vector>

// NOT THREAD SAFE
// provide external locking

// buffer size for a finest granularity "recv" operation
#define SOCKET_RECV_BUFFER	4096

class Socket
{
	public:
		Socket(const Socket&);
		virtual Socket& operator= (const Socket&);
		virtual ~Socket();

		virtual bool operator== (const Socket&);
		virtual bool server() = 0;

		int get_sock();
		bool valid() { return _sock != -1; }

		bool nonblocking(); // return whether O_NONBLOCK is set
		bool nonblocking(bool mode); // set O_NONBLOCK, return old state

		bool nondelaying(); // return whether TCP_NODELAY is set
		bool nondelaying(bool mode); // set TCP_NODELAY, return old state

	protected:
		Socket(int sock); // takes ownership of sock
		int _sock;
		void close();
		counting_auto_ptr<int> _counter;

		void poll(bool& read, bool& write, int timeout); // milliseconds

	private:
		void decrease_counter();
};

class ServerSocket;

class ClientSocket : public Socket
{
	public:
		ClientSocket();

		// UNIX domain socket
		ClientSocket(const String& sock_path);

		ClientSocket(	const String& hostname,
						unsigned short port,
						unsigned int timeout_ms=0);
		ClientSocket(const ClientSocket&);
		virtual ClientSocket& operator= (const ClientSocket&);
		virtual ~ClientSocket();

		/*
		** Try to receive SOCKET_RECV_BUFFER bytes and return them as String.
		**
		** In case of a nonblocking socket, this method expects that _sock is
		** successfully poll'd for POLLIN event prior to being called,
		** otherwise its behavior can be unexpected (specifically, socket
		** can be closed prematurely upon immediate EAGAIN/EWOULDBLOCK).
		**
		** Return value:
		** ret.length() == SOCKET_RECV_BUFFER
		**                the whole buffer received
		** ret.length() < SOCKET_RECV_BUFFER
		**                only the part of the buffer received due to EOF
		**                or possibly EAGAIN/EWOULDBLOCK if the contained
		**                socket is set as nonblocking
		**                (subsequent recv/poll will decide)
		**
		** Note: ret.length() == 0 (ret=="") never occurs; if EOF is apparent,
		**       contained socket is closed and the exception is raised.
		**
		** Upon non-recoverable condition (incl. closing the socket, see note),
		** exception is raised.  Possible interruptions upon signal receipt are
		** handled internally.
		*/
		virtual String recv();

		/*
		** Timeout-version of @recv.
		**
		** If timeout passes by, returns "".
		*/
		virtual String recv(int timeout);

		/*
		** Try to send @msg.length() bytes of @msg and return what is left.
		**
		** Return value:
		** ""             the whole @msg has been sent or @msg == ""
		** ret.length() < @msg.length()
		**                only the part of @msg has been sent due to possibly
		**                EAGAIN/EWOULDBLOCK if the contained socket is set
		**                as nonblocking
		**
		** Note: ret == @msg shall not ever occur upon successfully polling
		**       for POLLOUT event beforehand (unless @msg == "").
		**
		** Upon non-recoverable condition, exception is raised.  Possible
		** interruptions upon signal receipt are handled internally.
		*/
		virtual String send(const String& msg);

		/*
		** Timeout-version of @send.
		**
		** If timeout passes by, returns @msg.  In case of a nonblocking
		** socket, this is distinguishable only when polling for POLLOUT
		** event beforehand, see note by the base version of send().
		*/
		virtual String send(const String& msg, int timeout);

		virtual void ready(bool& recv, bool& send, int timeout);
		virtual bool server() { return false; }
		virtual bool connected_to(const String& hostname);

	protected:
		// address in network byte order
		u_int32_t _addr;

		// takes ownership of sock
		ClientSocket(int sock, u_int32_t addr=0);

	friend class ServerSocket;
};

class ServerSocket : public Socket
{
	public:
		ServerSocket(const String& sock_path); // UNIX socket
		ServerSocket(unsigned short port); // TCP socket
		ServerSocket(const ServerSocket&);
		virtual ServerSocket& operator= (const ServerSocket&);
		virtual ~ServerSocket();

		ClientSocket accept();

		virtual bool ready(int timeout);

		virtual bool server() { return true; }

	private:
		bool _unix_sock;
		String _sock_path;
};

#endif
